/*
 * softGL - A software based graphics library for PS3
 * -----------------------------------------------------------------------
 * Licensed under the BSD license, see LICENSE in PS3SDK root for details.
 *
 * gfx.c - Graphics Functions
 *
 * Copyright (c) 2010 Ifcaro <taslokooo@gmail.com>
 */
 
#include <ps3hv.h>
#include <ps3tb.h>
#include <malloc.h>
#include "gfx.h"
#include "math.h"

#define L1GPU_CONTEXT_ATTRIBUTE_DISPLAY_FLIP    0x102

static gpu_memory_handle_t memory_handle;
uint64_t context_handle, lpar_dma_control, lpar_driver_info, lpar_reports, lpar_reports_size;
uint64_t vram_base, fb_size, fb_frame;

uint32_t ticks;

// System

void InitGFX(uint32_t mode){
    ps3AvVideoInit();
    ps3AvAudioInit();
    ps3AvAvInit(0);
    ps3AvAvbParamDualVideo(mode, mode);
	
	screen.width=modeTab[mode].width;
	screen.height=modeTab[mode].height;

    ps3GpuInit();
	
    ps3MmuInit();
    ps3MmuOn();
    ps3ConsoleInit();
	
	memory_handle=(gpu_memory_handle_t)ps3GpuGetVramMem();
	vram_base=(uint64_t)ps3GpuGetVramBase();
	fb_size = screen.width*screen.height*4;
	screen.ptr=(void*)vram_base;
	
	lv1_gpu_context_allocate(memory_handle, 0, &context_handle, &lpar_dma_control, &lpar_driver_info, &lpar_reports, &lpar_reports_size);
	
	fb_frame=0;
}

void Flip(){
	lv1_gpu_context_attribute(context_handle, L1GPU_CONTEXT_ATTRIBUTE_DISPLAY_FLIP, 0, vram_base+(fb_size*fb_frame), 0, 0); // Fix flickering in hdmi mode
	lv1_gpu_context_attribute(context_handle, L1GPU_CONTEXT_ATTRIBUTE_DISPLAY_FLIP, 1, vram_base+(fb_size*fb_frame), 0, 0);
	if(fb_frame<3){
		fb_frame++;
	}else{
		fb_frame=0;
	}
	screen.ptr=(uint32_t*)(vram_base+(fb_size*fb_frame));

}

uint32_t GetTicks(){
	if(ticks<4294967295){
		ticks++;
	}else{
		ticks=0;
	}

	return ticks;
}

// Loading functions

TFont *LoadFont(TImage *image){
	TFont *fuente = NF_New();
	NF_Push(fuente);

	NF_ResetFont(&screen, image, 1);

	return fuente;
}

TImage *LoadImageRaw(void *src, int width, int height){
	TImage *imagen=(TImage*)malloc(sizeof(TImage));

	imagen->width=width;
	imagen->height=height;
	imagen->ptr=src;

	return imagen;
}

// Primitives

void DrawPoint(uint32_t color, int px, int py){
	uint32_t *fb=(uint32_t *)screen.ptr;

	if(px>0 && py>0 && px<screen.width && py<screen.height){
		int alpha = (color&0xff);
		color >>= 8;
		if(!alpha){
		
		}else if(alpha == 0xff){
			fb[screen.width*py+px] = color;
		}else{
			uint32_t v2 = fb[screen.width*py+px];
			int rb = ((color&0xff00ff)*alpha + (v2&0xff00ff)*(255-alpha))>>8;
			int g = ((color&0x00ff00)*alpha + (v2&0x00ff00)*(255-alpha))>>8;
			fb[screen.width*py+px] = (rb & 0xff00ff)|(g & 0x00ff00);
		}
	}
}

void DrawLine(uint32_t color, int x0, int y0, int x1, int y1){
        int dy = y1 - y0;
        int dx = x1 - x0;
        int stepx, stepy;

        if (dy < 0) { dy = -dy;  stepy = -1; } else { stepy = 1; }
        if (dx < 0) { dx = -dx;  stepx = -1; } else { stepx = 1; }
        dy <<= 1;                                                  // dy is now 2*dy
        dx <<= 1;                                                  // dx is now 2*dx

        DrawPoint(color, x0, y0);
        if (dx > dy) {
            int fraction = dy - (dx >> 1);                         // same as 2*dy - dx
            while (x0 != x1) {
                if (fraction >= 0) {
                    y0 += stepy;
                    fraction -= dx;                                // same as fraction -= 2*dx
                }
                x0 += stepx;
                fraction += dy;                                    // same as fraction -= 2*dy
                DrawPoint(color, x0, y0);
            }
        } else {
            int fraction = dx - (dy >> 1);
            while (y0 != y1) {
                if (fraction >= 0) {
                    x0 += stepx;
                    fraction -= dy;
                }
                y0 += stepy;
                fraction += dx;
                DrawPoint(color, x0, y0);
            }
        }
    }

void DrawRect(uint32_t color, int px, int py, int w, int h){
	int x, y;

	for(y=0; y<h; y++) {
		for(x=0; x<w; x++) {
			DrawPoint(color, px+x, py+y);
		}
	}  
}

void DrawCircle(uint32_t color, int xCenter, int yCenter, int radius){
    int x, y, r2;

    r2 = radius * radius;
    DrawPoint(color, xCenter, yCenter + radius);
    DrawPoint(color, xCenter, yCenter - radius);
    DrawPoint(color, xCenter + radius, yCenter);
    DrawPoint(color, xCenter - radius, yCenter);

    y = radius;
    x = 1;
    y = (int) (sqrtf(r2 - 1) + 0.5);
    while (x < y) {
        DrawPoint(color, xCenter + x, yCenter + y);
        DrawPoint(color, xCenter + x, yCenter - y);
        DrawPoint(color, xCenter - x, yCenter + y);
        DrawPoint(color, xCenter - x, yCenter - y);
        DrawPoint(color, xCenter + y, yCenter + x);
        DrawPoint(color, xCenter + y, yCenter - x);
        DrawPoint(color, xCenter - y, yCenter + x);
        DrawPoint(color, xCenter - y, yCenter - x);
        x += 1;
        y = (int) (sqrtf(r2 - x*x) + 0.5);
    }
    if (x == y) {
        DrawPoint(color, xCenter + x, yCenter + y);
        DrawPoint(color, xCenter + x, yCenter - y);
        DrawPoint(color, xCenter - x, yCenter + y);
        DrawPoint(color, xCenter - x, yCenter - y);
    }
}

void DrawCircleFill(uint32_t color, int xCenter, int yCenter, int radius){
    int x, y, r2;

    r2 = radius * radius;
    DrawLine(color, xCenter, yCenter, xCenter + radius, yCenter);
    DrawLine(color, xCenter, yCenter, xCenter - radius, yCenter);

    y = radius;
    x = 1;
    y = (int) (sqrtf(r2 - 1) + 0.5);
    while (x < y) {
		DrawLine(color, xCenter, yCenter + y, xCenter + x, yCenter + y);
		DrawLine(color, xCenter, yCenter - y, xCenter + x, yCenter - y);
		DrawLine(color, xCenter, yCenter + y, xCenter - x, yCenter + y);
		DrawLine(color, xCenter, yCenter - y, xCenter - x, yCenter - y);
		DrawLine(color, xCenter, yCenter + x, xCenter + y, yCenter + x);
		DrawLine(color, xCenter, yCenter - x, xCenter + y, yCenter - x);
		DrawLine(color, xCenter, yCenter + x, xCenter - y, yCenter + x);
		DrawLine(color, xCenter, yCenter - x, xCenter - y, yCenter - x);
        x += 1;
        y = (int) (sqrtf(r2 - x*x) + 0.5);
    }
    if (x == y) {
        DrawLine(color, xCenter, yCenter + y, xCenter + x, yCenter + y);
        DrawLine(color, xCenter, yCenter - y, xCenter + x, yCenter - y);
        DrawLine(color, xCenter, yCenter + y, xCenter - x, yCenter + y);
        DrawLine(color, xCenter, yCenter - y, xCenter - x, yCenter - y);
    }
}

void Cls(uint32_t color){
	int x, y;
	uint32_t *fb=(uint32_t *)screen.ptr;

	for(y=0;y<screen.height;y++){
		for(x=0;x<screen.width;x++){
			fb[screen.width*y+x]=color;
		}
	}
}

// Images

void DrawImage(TImage *image, int px, int py, int w, int h){
	int x, y;
	uint32_t *pdata=(uint32_t*)image->ptr;

	float scalex=(float)image->width/(float)w;
	float scaley=(float)image->height/(float)h;

	for(y=0; y<h; y++) {
		for(x=0; x<w; x++) {
			uint32_t v = pdata[(image->width*(int)(y*scaley))+(int)(x*scalex)];
			DrawPoint(v, px+x, py+y);
		}
	}  
}
	
void BlitImage(TImage *imgsrc, TRect *srcrect, TImage *imgdst, TRect *dstrect){
	int x, y;
	uint32_t *src=(uint32_t *)imgsrc->ptr;
	uint32_t *dst=(uint32_t *)imgdst->ptr;

	for(y=0;y<srcrect->h;y++){
		for(x=0;x<srcrect->w;x++){
			if((uint64_t)imgdst->ptr==(uint64_t)screen.ptr){
				DrawPoint(src[imgsrc->width*(srcrect->y+y)+(srcrect->x+x)],dstrect->x+x,dstrect->y+y);
			}else if(dstrect->x+x<imgdst->width && dstrect->y+y<imgdst->height){
				dst[imgdst->width*(dstrect->y+y)+(dstrect->x+x)]=src[imgsrc->width*(srcrect->y+y)+(srcrect->x+x)];
			}
		}
	}
}

uint32_t GetPixel(TImage *image, int x, int y){
	uint32_t *fb=(uint32_t *)image->ptr;

	return fb[image->width*y+x];
}

